@@ -129,7 +129,7 @@ def get_nearest_node(G, point, method="haversine", return_dist=False):
         is the distance (in meters if haversine, or graph node coordinate
         units if euclidean) between the point and nearest node
     """
-    if len(G) < 1:
+    if not G:
         raise ValueError("G must contain at least one node")
 
     # dump graph node coordinates into a pandas dataframe indexed by node id
@@ -169,6 +169,8 @@ def get_nearest_edge(G, point, return_geom=False, return_dist=False):
     """
     Find the nearest edge to a point by minimum Euclidean distance.
 
+    For best results, both G and point should be projected.
+
     Parameters
     ----------
     G : networkx.MultiDiGraph
@@ -190,18 +192,16 @@ def get_nearest_edge(G, point, return_geom=False, return_dist=False):
         Or a tuple of (u, v, key, dist) if return_dist is True.
         Or a tuple of (u, v, key, geom, dist) if return_geom and return_dist are True.
     """
-    # get u, v, key, geom from all the graph edges
-    gdf_edges = utils_graph.graph_to_gdfs(G, nodes=False, fill_edge_geometry=True)
-    edges = gdf_edges[["u", "v", "key", "geometry"]].values
-
     # convert lat,lng (y,x) point to x,y for shapely distance operation
     xy_point = Point(reversed(point))
 
     # calculate euclidean distance from each edge's geometry to this point
-    edge_distances = [(edge, xy_point.distance(edge[3])) for edge in edges]
+    gs_edges = utils_graph.graph_to_gdfs(G, nodes=False)["geometry"]
+    uvk_geoms = zip(gs_edges.index, gs_edges.values)
+    distances = ((uvk, geom, xy_point.distance(geom)) for uvk, geom in uvk_geoms)
 
     # the nearest edge minimizes the distance to the point
-    (u, v, key, geom), dist = min(edge_distances, key=lambda x: x[1])
+    (u, v, key), geom, dist = min(distances, key=lambda x: x[2])
     utils.log(f"Found nearest edge ({u, v, key}) to point {point}")
 
     # return results requested by caller
@@ -258,8 +258,8 @@ def get_nearest_nodes(G, X, Y, method=None):
     elif method == "kdtree":
 
         # check if we were able to import scipy.spatial.cKDTree
-        if not cKDTree:
-            raise ImportError("The scipy package must be installed to use this optional feature.")
+        if cKDTree is None:
+            raise ImportError("scipy must be installed to use this optional feature")
 
         # build a k-d tree for euclidean nearest node search
         nodes = pd.DataFrame(
@@ -275,10 +275,8 @@ def get_nearest_nodes(G, X, Y, method=None):
     elif method == "balltree":
 
         # check if we were able to import sklearn.neighbors.BallTree
-        if not BallTree:
-            raise ImportError(
-                "The scikit-learn package must be installed to use this optional feature."
-            )
+        if BallTree is None:
+            raise ImportError("scikit-learn must be installed to use this optional feature")
 
         # haversine requires data in form of [lat, lng] and inputs/outputs in
         # units of radians
@@ -361,11 +359,11 @@ def get_nearest_edges(G, X, Y, method=None, dist=0.0001):
     elif method == "kdtree":
 
         # check if we were able to import scipy.spatial.cKDTree successfully
-        if not cKDTree:
-            raise ImportError("The scipy package must be installed to use this optional feature.")
+        if cKDTree is None:
+            raise ImportError("scipy must be installed to use this optional feature")
 
         # transform graph into DataFrame
-        edges = utils_graph.graph_to_gdfs(G, nodes=False, fill_edge_geometry=True)
+        edges = utils_graph.graph_to_gdfs(G, nodes=False).reset_index()
 
         # transform edges into evenly spaced points
         edges["points"] = edges.apply(
@@ -403,13 +401,11 @@ def get_nearest_edges(G, X, Y, method=None, dist=0.0001):
     elif method == "balltree":
 
         # check if we were able to import sklearn.neighbors.BallTree successfully
-        if not BallTree:
-            raise ImportError(
-                "The scikit-learn package must be installed to use this optional feature."
-            )
+        if BallTree is None:
+            raise ImportError("scikit-learn must be installed to use this optional feature")
 
         # transform graph into DataFrame
-        edges = utils_graph.graph_to_gdfs(G, nodes=False, fill_edge_geometry=True)
+        edges = utils_graph.graph_to_gdfs(G, nodes=False).reset_index()
 
         # transform edges into evenly spaced points
         edges["points"] = edges.apply(
