@@ -1,16 +1,14 @@
 """Interact with the OSM APIs."""
 
-import bz2
 import datetime as dt
 import json
 import logging as lg
 import math
-import os
 import re
 import time
-import xml.sax
 from collections import OrderedDict
-from hashlib import md5
+from hashlib import sha1
+from pathlib import Path
 
 import requests
 from dateutil import parser as date_parser
@@ -19,6 +17,7 @@ from . import projection
 from . import settings
 from . import utils
 from . import utils_geo
+from ._errors import CacheOnlyModeInterrupt
 
 
 def _get_osm_filter(network_type):
@@ -27,9 +26,8 @@ def _get_osm_filter(network_type):
 
     Parameters
     ----------
-    network_type : string
-        {'walk', 'bike', 'drive', 'drive_service', 'all', 'all_private'}
-        what type of street or other network to get
+    network_type : string {"all_private", "all", "bike", "drive", "drive_service", "walk"}
+        what type of street network to get
 
     Returns
     -------
@@ -138,19 +136,16 @@ def _save_to_cache(url, response_json, sc):
 
         else:
             # create the folder on the disk if it doesn't already exist
-            if not os.path.exists(settings.cache_folder):
-                os.makedirs(settings.cache_folder)
+            cache_folder = Path(settings.cache_folder)
+            cache_folder.mkdir(parents=True, exist_ok=True)
 
             # hash the url to make the filename succinct but unique
-            # md5 digest is 128 bits = 16 bytes = 32 hexadecimal characters
-            filename = md5(url.encode("utf-8")).hexdigest()
-            cache_filepath = os.path.join(settings.cache_folder, os.extsep.join([filename, "json"]))
+            # sha1 digest is 160 bits = 20 bytes = 40 hexadecimal characters
+            filename = sha1(url.encode("utf-8")).hexdigest() + ".json"
+            cache_filepath = cache_folder / filename
 
             # dump to json, and save to file
-            json_str = str(json.dumps(response_json))
-            with open(cache_filepath, "w", encoding="utf-8") as cache_file:
-                cache_file.write(json_str)
-
+            cache_filepath.write_text(json.dumps(response_json), encoding="utf-8")
             utils.log(f'Saved response to cache file "{cache_filepath}"')
 
 
@@ -167,18 +162,15 @@ def _url_in_cache(url):
 
     Returns
     -------
-    filepath : string
+    filepath : pathlib.Path
         path to cached response for url if it exists, otherwise None
     """
     # hash the url to generate the cache filename
-    filename = md5(url.encode("utf-8")).hexdigest()
-    filepath = os.path.join(settings.cache_folder, os.extsep.join([filename, "json"]))
+    filename = sha1(url.encode("utf-8")).hexdigest() + ".json"
+    filepath = Path(settings.cache_folder) / filename
 
     # if this file exists in the cache, return its full path
-    if os.path.isfile(filepath):
-        return filepath
-    else:
-        return None
+    return filepath if filepath.is_file() else None
 
 
 def _retrieve_from_cache(url, check_remark=False):
@@ -204,14 +196,13 @@ def _retrieve_from_cache(url, check_remark=False):
         # return cached response for this url if exists, otherwise return None
         cache_filepath = _url_in_cache(url)
         if cache_filepath is not None:
-            with open(cache_filepath, encoding="utf-8") as cache_file:
-                response_json = json.load(cache_file)
+            response_json = json.loads(cache_filepath.read_text(encoding="utf-8"))
 
-                # return None if check_remark is True and there is a server
-                # remark in the cached response
-                if check_remark and "remark" in response_json:
-                    utils.log(f'Found remark, so ignoring cache file "{cache_filepath}"')
-                    return None
+            # return None if check_remark is True and there is a server
+            # remark in the cached response
+            if check_remark and "remark" in response_json:
+                utils.log(f'Found remark, so ignoring cache file "{cache_filepath}"')
+                return None
 
             utils.log(f'Retrieved response from cache file "{cache_filepath}"')
             return response_json
@@ -274,7 +265,7 @@ def _get_pause(recursive_delay=5, default_duration=60):
         status = response.text.split("\n")[3]
         status_first_token = status.split(" ")[0]
 
-    except Exception:
+    except Exception:  # pragma: no cover
         # if we cannot reach the status endpoint or parse its output, log an
         # error and return default duration
         sc = response.status_code
@@ -359,7 +350,7 @@ def _create_overpass_query(polygon_coord_str, tags):
     polygon_coord_str : list
         list of lat lng coordinates
     tags : dict
-        dict of tags used for finding geometry in the selected area
+        dict of tags used for finding elements in the selected area
 
     Returns
     -------
@@ -421,22 +412,23 @@ def _create_overpass_query(polygon_coord_str, tags):
     return query
 
 
-def _osm_net_download(polygon, network_type, custom_filter):
+def _osm_network_download(polygon, network_type, custom_filter):
     """
-    Download OSM ways and nodes within some polygon from the Overpass API.
+    Retrieve networked ways and nodes within boundary from the Overpass API.
 
     Parameters
     ----------
     polygon : shapely.geometry.Polygon or shapely.geometry.MultiPolygon
-        geographic boundaries to fetch the street network within
+        boundary to fetch the network ways/nodes within
     network_type : string
-        what type of street network to get if custom_filter is not None
+        what type of street network to get if custom_filter is None
     custom_filter : string
-        a custom network filter to be used instead of the network_type presets
+        a custom ways filter to be used instead of the network_type presets
 
     Returns
     -------
     response_jsons : list
+        list of JSON responses from the Overpass server
     """
     # create a filter to exclude certain kinds of ways based on the requested
     # network_type, if provided, otherwise use custom_filter
@@ -463,22 +455,22 @@ def _osm_net_download(polygon, network_type, custom_filter):
         f"Got all network data within polygon from API in {len(polygon_coord_strs)} request(s)"
     )
 
+    if settings.cache_only_mode:
+        raise CacheOnlyModeInterrupt("settings.cache_only_mode=True")
+
     return response_jsons
 
 
-def _osm_geometry_download(polygon, tags):
+def _osm_geometries_download(polygon, tags):
     """
-    Download all OSM geometry within some polygon from the Overpass API.
-
-    Note that if a polygon is passed-in, the query will be limited to the
-    exterior ring only.
+    Retrieve non-networked elements within boundary from the Overpass API.
 
     Parameters
     ----------
     polygon : shapely.geometry.Polygon
-        geographic boundaries to fetch geometry within
+        boundaries to fetch elements within
     tags : dict
-        dict of tags used for finding geometry in the selected area
+        dict of tags used for finding elements in the selected area
 
     Returns
     -------
@@ -497,51 +489,62 @@ def _osm_geometry_download(polygon, tags):
         response_jsons.append(response_json)
 
     utils.log(
-        f"Got all geometry data within polygon from API in {len(polygon_coord_strs)} request(s)"
+        f"Got all geometries data within polygon from API in {len(polygon_coord_strs)} request(s)"
     )
 
     return response_jsons
 
 
-def _osm_polygon_download(query, limit=1, polygon_geojson=1):
+def _osm_place_download(query, by_osmid=False, limit=1, polygon_geojson=1):
     """
-    Geocode a place and download its boundary geometry from Nominatim API.
+    Retrieve a place from the Nominatim API.
 
     Parameters
     ----------
     query : string or dict
-        query string or structured query dict to geocode/download
+        query string or structured query dict
+    by_osmid : bool
+        if True, handle query as an OSM ID for lookup rather than text search
     limit : int
         max number of results to return
     polygon_geojson : int
-        request the boundary geometry polygon from the API, 0=no, 1=yes
+        retrieve the place's geometry from the API, 0=no, 1=yes
 
     Returns
     -------
     response_json : dict
+        JSON response from the Nominatim server
     """
     # define the parameters
     params = OrderedDict()
     params["format"] = "json"
-    params["limit"] = limit
-    # prevent OSM from deduping results so we get precisely 'limit' # of results
-    params["dedupe"] = 0
     params["polygon_geojson"] = polygon_geojson
 
-    # add the structured query dict (if provided) to params, otherwise query
-    # with place name string
-    if isinstance(query, str):
-        params["q"] = query
-    elif isinstance(query, dict):
-        # add the query keys in alphabetical order so the URL is the same string
-        # each time, for caching purposes
-        for key in sorted(list(query.keys())):
-            params[key] = query[key]
+    if by_osmid:
+        # if querying by OSM ID, use the lookup endpoint
+        request_type = "lookup"
+        params["osm_ids"] = query
+
     else:
-        raise TypeError("query must be a dict or a string")
+        # if not querying by OSM ID, use the search endpoint
+        request_type = "search"
+
+        # prevent OSM from deduping so we get precise number of results
+        params["dedupe"] = 0
+        params["limit"] = limit
+
+        if isinstance(query, str):
+            params["q"] = query
+        elif isinstance(query, dict):
+            # add query keys in alphabetical order so URL is the same string
+            # each time, for caching purposes
+            for key in sorted(query):
+                params[key] = query[key]
+        else:
+            raise TypeError("query must be a dict or a string")
 
     # request the URL, return the JSON
-    response_json = nominatim_request(params=params)
+    response_json = nominatim_request(params=params, request_type=request_type)
     return response_json
 
 
@@ -553,8 +556,8 @@ def nominatim_request(params, request_type="search", pause=1, error_pause=60):
     ----------
     params : OrderedDict
         key-value pairs of parameters
-    request_type : string
-        Type of Nominatim query. One of: search, reverse, or lookup
+    request_type : string {"search", "reverse", "lookup"}
+        which Nominatim API endpoint to query
     pause : int
         how long to pause before request, in seconds. per the nominatim usage
         policy: "an absolute maximum of 1 request per second" is allowed
@@ -686,85 +689,3 @@ def overpass_request(data, pause=None, error_pause=60):
 
         _save_to_cache(prepared_url, response_json, sc)
         return response_json
-
-
-def _overpass_json_from_file(filepath):
-    """
-    Read OSM XML from file and return Overpass-like JSON.
-
-    Parameters
-    ----------
-    filepath : string
-        path to file containing OSM XML data
-
-    Returns
-    -------
-    OSMContentHandler object
-    """
-
-    def _opener(filepath):
-        _, ext = os.path.splitext(filepath)
-        if ext == ".bz2":
-            return bz2.BZ2File(filepath)
-        else:
-            # assume an unrecognized file extension is just XML
-            return open(filepath, mode="rb")
-
-    with _opener(filepath) as file:
-        handler = _OSMContentHandler()
-        xml.sax.parse(file, handler)
-        return handler.object
-
-
-class _OSMContentHandler(xml.sax.handler.ContentHandler):
-    """
-    SAX content handler for OSM XML.
-
-    Used to build an Overpass-like response JSON object in self.object. For format
-    notes, see http://wiki.openstreetmap.org/wiki/OSM_XML#OSM_XML_file_format_notes
-    and http://overpass-api.de/output_formats.html#json
-    """
-
-    def __init__(self):
-        self._element = None
-        self.object = {"elements": []}
-
-    def startElement(self, name, attrs):
-        if name == "osm":
-            self.object.update({k: attrs[k] for k in attrs.keys() if k in {"version", "generator"}})
-
-        elif name in {"node", "way"}:
-            self._element = dict(type=name, tags={}, nodes=[], **attrs)
-            self._element.update({k: float(attrs[k]) for k in attrs.keys() if k in {"lat", "lon"}})
-            self._element.update(
-                {
-                    k: int(attrs[k])
-                    for k in attrs.keys()
-                    if k in {"id", "uid", "version", "changeset"}
-                }
-            )
-
-        elif name == "relation":
-            self._element = dict(type=name, tags={}, members=[], **attrs)
-            self._element.update(
-                {
-                    k: int(attrs[k])
-                    for k in attrs.keys()
-                    if k in {"id", "uid", "version", "changeset"}
-                }
-            )
-
-        elif name == "tag":
-            self._element["tags"].update({attrs["k"]: attrs["v"]})
-
-        elif name == "nd":
-            self._element["nodes"].append(int(attrs["ref"]))
-
-        elif name == "member":
-            self._element["members"].append(
-                {k: (int(attrs[k]) if k == "ref" else attrs[k]) for k in attrs.keys()}
-            )
-
-    def endElement(self, name):
-        if name in {"node", "way", "relation"}:
-            self.object["elements"].append(self._element)
