@@ -1,6 +1,6 @@
 """Plot spatial geometries, street networks, and routes."""
 
-import os
+from pathlib import Path
 
 import matplotlib.cm as cm
 import matplotlib.colors as colors
@@ -19,7 +19,7 @@ from . import utils_graph
 
 def get_colors(n, cmap="viridis", start=0.0, stop=1.0, alpha=1.0, return_hex=False):
     """
-    Get n evenly-spaced colors from a matplotlib colormap.
+    Get `n` evenly-spaced colors from a matplotlib colormap.
 
     Parameters
     ----------
@@ -115,7 +115,7 @@ def get_edge_colors_by_attr(
     Returns
     -------
     edge_colors : pandas.Series
-        series labels are edge IDs (u, v, k) and values are colors
+        series labels are edge IDs (u, v, key) and values are colors
     """
     vals = pd.Series(nx.get_edge_attributes(G, attr))
     return _get_colors_by_value(vals, num_bins, cmap, start, stop, na_color, equal_size)
@@ -339,7 +339,7 @@ def plot_graph_routes(G, routes, route_colors="r", **pgr_kwargs):
         matplotlib figure, axis
     """
     # check for valid arguments
-    if not all([isinstance(r, list) for r in routes]):
+    if not all(isinstance(r, list) for r in routes):
         raise ValueError("routes must be a list of route lists")
     if len(routes) < 2:
         raise ValueError("You must pass more than 1 route")
@@ -410,7 +410,7 @@ def plot_figure_ground(
     dist : numeric
         how many meters to extend north, south, east, west from center point
     network_type : string
-        what type of network to get
+        what type of street network to get
     street_widths : dict
         dict keys are street types and values are widths to plot in pixels
     default_width : numeric
@@ -492,11 +492,11 @@ def plot_figure_ground(
     if smooth_joints:
         # for each node, get a nodesize according to the narrowest incident edge
         node_widths = dict()
-        for node in Gu.nodes():
+        for node in Gu.nodes:
             # first, identify all the highway types of this node's incident edges
             ie_data = [Gu.get_edge_data(node, nbr) for nbr in Gu.neighbors(node)]
             edge_types = [d[min(d)]["highway"] for d in ie_data]
-            if len(edge_types) < 1:
+            if not edge_types:
                 # if node has no incident edges, make size zero
                 node_widths[node] = 0
             else:
@@ -522,7 +522,7 @@ def plot_figure_ground(
                 node_widths[node] = circle_area
 
         # assign the node size to each node in the graph
-        node_sizes = [node_widths[node] for node in Gu.nodes()]
+        node_sizes = [node_widths[node] for node in Gu.nodes]
     else:
         node_sizes = 0
 
@@ -550,6 +550,7 @@ def plot_footprints(
     ax=None,
     figsize=(8, 8),
     color="orange",
+    alpha=None,
     bgcolor="#111111",
     bbox=None,
     save=False,
@@ -571,6 +572,8 @@ def plot_footprints(
         if ax is None, create new figure with size (width, height)
     color : string
         color of the footprints
+    alpha : float
+        opacity of the footprints
     bgcolor : string
         background color of the plot
     bbox : tuple
@@ -601,7 +604,7 @@ def plot_footprints(
 
     # retain only Polygons and MultiPolygons, then plot
     gdf = gdf[gdf["geometry"].type.isin({"Polygon", "MultiPolygon"})]
-    ax = gdf.plot(ax=ax, facecolor=color, edgecolor="none", linewidth=0, alpha=1)
+    ax = gdf.plot(ax=ax, facecolor=color, edgecolor="none", linewidth=0, alpha=alpha)
 
     # determine figure extents
     if bbox is None:
@@ -705,16 +708,15 @@ def _save_and_show(fig, ax, save=False, show=True, close=True, filepath=None, dp
 
         # default filepath, if none provided
         if filepath is None:
-            filepath = os.path.join(settings.imgs_folder, "image.png")
+            filepath = Path(settings.imgs_folder) / "image.png"
+        else:
+            filepath = Path(filepath)
 
         # if save folder does not already exist, create it
-        folder, _ = os.path.split(filepath)
-        if not folder == "" and not os.path.exists(folder):
-            os.makedirs(folder)
+        filepath.parent.mkdir(parents=True, exist_ok=True)
 
         # get the file extension and figure facecolor
-        _, ext = os.path.splitext(filepath)
-        ext = ext.strip(".")
+        ext = filepath.suffix.strip(".")
         fc = fig.get_facecolor()
 
         if ext == "svg":
