@@ -4,9 +4,11 @@ import logging as lg
 
 import geopandas as gpd
 import networkx as nx
+import pandas as pd
 from shapely.geometry import LineString
 from shapely.geometry import Point
 from shapely.geometry import Polygon
+from shapely.ops import unary_union
 
 from . import utils
 from . import utils_graph
@@ -135,7 +137,7 @@ def _build_path(G, endpoint, endpoint_successor, endpoints):
                         # we have come to the end of a self-looping edge, so
                         # add first node to end of path to close it and return
                         return path + [endpoint]
-                    else:
+                    else:  # pragma: no cover
                         # this can happen due to OSM digitization error where
                         # a one-way street turns into a two-way here, but
                         # duplicate incoming one-way edges are present
@@ -143,7 +145,7 @@ def _build_path(G, endpoint, endpoint_successor, endpoints):
                             f"Unexpected simplify pattern handled near {successor}", level=lg.WARN
                         )
                         return path
-                else:
+                else:  # pragma: no cover
                     # if successor has >1 successors, then successor must have
                     # been an endpoint because you can go in 2 new directions.
                     # this should never occur in practice
@@ -177,45 +179,30 @@ def _get_paths_to_simplify(G, strict=True):
     path_to_simplify : list
     """
     # first identify all the nodes that are endpoints
-    endpoints = set([n for n in G.nodes() if _is_endpoint(G, n, strict=strict)])
+    endpoints = set([n for n in G.nodes if _is_endpoint(G, n, strict=strict)])
     utils.log(f"Identified {len(endpoints)} edge endpoints")
 
     # for each endpoint node, look at each of its successor nodes
     for endpoint in endpoints:
         for successor in G.successors(endpoint):
             if successor not in endpoints:
-                # if endpoint node's successor is not an endpoint, build a path
+                # if endpoint node's successor is not an endpoint, build path
                 # from the endpoint node, through the successor, and on to the
                 # next endpoint node
                 yield _build_path(G, endpoint, successor, endpoints)
 
 
-def _is_simplified(G):
-    """
-    Determine if a graph has already had its topology simplified.
-
-    Parameters
-    ----------
-    G : networkx.MultiDiGraph
-        input graph
-
-    Returns
-    -------
-    bool
-    """
-    return "simplified" in G.graph and G.graph["simplified"]
-
-
 def simplify_graph(G, strict=True, remove_rings=True):
     """
     Simplify a graph's topology by removing interstitial nodes.
 
-    Simplify graph topology by removing all nodes that are not intersections
+    Simplifies graph topology by removing all nodes that are not intersections
     or dead-ends. Create an edge directly between the end points that
     encapsulate them, but retain the geometry of the original edges, saved as
-    an attribute in new edge. Some of the resulting consolidated edges may
-    comprise multiple OSM ways, and if so, their multiple attribute values are
-    stored as a list.
+    a new `geometry` attribute on the new edge. Note that only simplified
+    edges receive a `geometry` attribute. Some of the resulting consolidated
+    edges may comprise multiple OSM ways, and if so, their multiple attribute
+    values are stored as a list.
 
     Parameters
     ----------
@@ -232,14 +219,15 @@ def simplify_graph(G, strict=True, remove_rings=True):
     Returns
     -------
     G : networkx.MultiDiGraph
-        topologically simplified graph
+        topologically simplified graph, with a new `geometry` attribute on
+        each simplified edge
     """
-    if _is_simplified(G):
+    if "simplified" in G.graph and G.graph["simplified"]:
         raise Exception("This graph has already been simplified, cannot simplify it again.")
 
     utils.log("Begin topologically simplifying the graph...")
 
-    # make a copy to not edit the original graph object the caller passed in
+    # make a copy to not mutate original graph object caller passed in
     G = G.copy()
     initial_node_count = len(G)
     initial_edge_count = len(G.edges)
@@ -257,7 +245,7 @@ def simplify_graph(G, strict=True, remove_rings=True):
             # there should rarely be multiple edges between interstitial nodes
             # usually happens if OSM has duplicate ways digitized for just one
             # street... we will keep only one of the edges (see below)
-            if not G.number_of_edges(u, v) == 1:
+            if G.number_of_edges(u, v) != 1:
                 utils.log(f"Found multiple edges between {u} and {v} when simplifying")
 
             # get edge between these nodes: if multiple edges exist between
@@ -275,11 +263,11 @@ def simplify_graph(G, strict=True, remove_rings=True):
 
         for key in edge_attributes:
             # don't touch the length attribute, we'll sum it at the end
-            if len(set(edge_attributes[key])) == 1 and not key == "length":
+            if len(set(edge_attributes[key])) == 1 and key != "length":
                 # if there's only 1 unique value in this attribute list,
                 # consolidate it to the single value (the zero-th)
                 edge_attributes[key] = edge_attributes[key][0]
-            elif not key == "length":
+            elif key != "length":
                 # otherwise, if there are multiple values, keep one of each value
                 edge_attributes[key] = list(set(edge_attributes[key]))
 
@@ -309,7 +297,7 @@ def simplify_graph(G, strict=True, remove_rings=True):
         wccs = nx.weakly_connected_components(G)
         nodes_in_rings = set()
         for wcc in wccs:
-            if all([not _is_endpoint(G, n) for n in wcc]):
+            if not any(_is_endpoint(G, n) for n in wcc):
                 nodes_in_rings.update(wcc)
         G.remove_nodes_from(nodes_in_rings)
 
@@ -383,44 +371,91 @@ def consolidate_intersections(
         returns GeoSeries of shapely Points representing the centroids of
         street intersections
     """
-    # if dead_ends is False, discard dead-end nodes to retain only intersections
+    # if dead_ends is False, discard dead-ends to retain only intersections
     if not dead_ends:
-        if "streets_per_node" in G.graph:
-            streets_per_node = G.graph["streets_per_node"]
-        else:
-            streets_per_node = utils_graph.count_streets_per_node(G)
+        spn = nx.get_node_attributes(G, "street_count")
+        if set(spn) != set(G.nodes):
+            utils.log("Graph nodes changed since `street_count`s were calculated", level=lg.WARN)
+        dead_end_nodes = [node for node, count in spn.items() if count <= 1]
 
-        dead_end_nodes = [node for node, count in streets_per_node.items() if count <= 1]
-
-        # make a copy to not edit the original graph object the caller passed in
+        # make a copy to not mutate original graph object caller passed in
         G = G.copy()
         G.remove_nodes_from(dead_end_nodes)
 
     if rebuild_graph:
-        if len(G) == 0 or len(G.edges) == 0:
+        if not G or not G.edges:
             # cannot rebuild a graph with no nodes or no edges, just return it
             return G
         else:
-            return _consolidate_intersections_rebuild_graph(
-                G=G, tolerance=tolerance, reconnect_edges=reconnect_edges
-            )
+            return _consolidate_intersections_rebuild_graph(G, tolerance, reconnect_edges)
 
     else:
-        crs = G.graph["crs"]
-        if len(G) == 0:
+        if not G:
             # if graph has no nodes, just return empty GeoSeries
-            return gpd.GeoSeries(crs=crs)
+            return gpd.GeoSeries(crs=G.graph["crs"])
         else:
-            # create nodes gdf, buffer to passed-in distance, merge overlaps
-            gdf_nodes = utils_graph.graph_to_gdfs(G, edges=False)
-            merged_nodes = gdf_nodes.buffer(tolerance).unary_union
-            if isinstance(merged_nodes, Polygon):
-                # if only a single node results, make iterable to convert to GeoSeries
-                merged_nodes = [merged_nodes]
+            # return the centroids of the merged intersection polygons
+            return _merge_nodes_geometric(G, tolerance).centroid
 
-            # get the centroids of the merged intersection polygons
-            intersection_centroids = gpd.GeoSeries(list(merged_nodes), crs=crs).centroid
-            return intersection_centroids
+
+def _merge_nodes_geometric(G, tolerance, chunk=True):
+    """
+    Geometrically merge nodes within some distance of each other.
+
+    If chunk=True, it sorts the nodes GeoSeries by geometry x and y values (to
+    make unary_union faster), then buffers by tolerance. Next it divides the
+    nodes GeoSeries into n-sized chunks, where n = the square root of the
+    number of nodes. Then it runs unary_union on each chunk, and then runs
+    unary_union on the resulting unary unions. This is much faster on large
+    graphs (n>100000) because of how unary_union's runtime scales with vertex
+    count. But chunk=False is usually faster on small and medium sized graphs.
+    This hacky method will hopefully be made obsolete when shapely becomes
+    vectorized by incorporating the pygeos codebase.
+
+    Parameters
+    ----------
+    G : networkx.MultiDiGraph
+        a projected graph
+    tolerance : float
+        nodes are buffered to this distance (in graph's geometry's units) and
+        subsequent overlaps are dissolved into a single polygon
+    chunk : bool
+        if True, divide nodes into geometrically sorted chunks to improve the
+        speed of unary_union operation by running it on each chunk and then
+        running it on the results of those runs
+
+    Returns
+    -------
+    merged_nodes : GeoSeries
+        the merged overlapping polygons of the buffered nodes
+    """
+    # yield n-sized chunks of list one at a time
+    def get_chunks(lst, n):
+        for i in range(0, len(lst), n):
+            yield lst[i : i + n]
+
+    gs_nodes = utils_graph.graph_to_gdfs(G, edges=False)["geometry"]
+
+    if not chunk:
+        # buffer nodes GeoSeries then get unary union to merge overlaps
+        merged_nodes = gs_nodes.buffer(tolerance).unary_union
+
+    else:
+        # sort nodes GeoSeries by geometry x/y vals to make unary_union faster
+        idx = pd.DataFrame((gs_nodes.x, gs_nodes.y)).T.sort_values([0, 1]).index
+
+        # buffer nodes by tolerance then generate n-sized chunks of nodes
+        n = int(len(G) ** 0.5)
+        chunks = get_chunks(gs_nodes.loc[idx].buffer(tolerance).values, n)
+
+        # unary_union each chunk then unary_union those unary unions
+        merged_nodes = unary_union([unary_union(chunk) for chunk in chunks])
+
+    # if only a single node results, make it iterable to convert to GeoSeries
+    if isinstance(merged_nodes, Polygon):
+        merged_nodes = [merged_nodes]
+
+    return gpd.GeoSeries(list(merged_nodes), crs=G.graph["crs"])
 
 
 def _consolidate_intersections_rebuild_graph(G, tolerance=10, reconnect_edges=True):
@@ -459,25 +494,18 @@ def _consolidate_intersections_rebuild_graph(G, tolerance=10, reconnect_edges=Tr
         edge geometries
     """
     # STEP 1
-    # buffer nodes to passed-in distance, merge overlaps
-    gdf_nodes = utils_graph.graph_to_gdfs(G, edges=False)
-    buffered_nodes = gdf_nodes.buffer(tolerance).unary_union
-    if isinstance(buffered_nodes, Polygon):
-        # if only a single node results, make iterable to convert to GeoSeries
-        buffered_nodes = [buffered_nodes]
-
-    # STEP 2
-    # attach each node to its cluster of merged nodes
-    # first get the original graph's node points
-    node_points = gdf_nodes[["geometry"]]
-
-    # then turn buffered nodes into gdf and get centroids of each cluster as x, y
-    node_clusters = gpd.GeoDataFrame(geometry=list(buffered_nodes), crs=node_points.crs)
+    # buffer nodes to passed-in distance and merge overlaps. turn merged nodes
+    # into gdf and get centroids of each cluster as x, y
+    node_clusters = gpd.GeoDataFrame(geometry=_merge_nodes_geometric(G, tolerance))
     centroids = node_clusters.centroid
     node_clusters["x"] = centroids.x
     node_clusters["y"] = centroids.y
 
-    # then spatial join to give each node the label of cluster it's within
+    # STEP 2
+    # attach each node to its cluster of merged nodes. first get the original
+    # graph's node points then spatial join to give each node the label of
+    # cluster it's within
+    node_points = utils_graph.graph_to_gdfs(G, edges=False)[["geometry"]]
     gdf = gpd.sjoin(node_points, node_clusters, how="left", op="within")
     gdf = gdf.drop(columns="geometry").rename(columns={"index_right": "cluster"})
 
@@ -499,7 +527,7 @@ def _consolidate_intersections_rebuild_graph(G, tolerance=10, reconnect_edges=Tr
                     subcluster_centroid = node_points.loc[wcc].unary_union.centroid
                     gdf.loc[wcc, "x"] = subcluster_centroid.x
                     gdf.loc[wcc, "y"] = subcluster_centroid.y
-                    # move to subcluster by appending suffix to nodes cluster label
+                    # move to subcluster by appending suffix to cluster label
                     gdf.loc[wcc, "cluster"] = f"{cluster_label}-{suffix}"
                     suffix += 1
 
@@ -528,14 +556,14 @@ def _consolidate_intersections_rebuild_graph(G, tolerance=10, reconnect_edges=Tr
                 y=nodes_subset["y"].iloc[0],
             )
 
-    if len(G.edges) == 0 or not reconnect_edges:
+    if not G.edges or not reconnect_edges:
         # if reconnect_edges is False or there are no edges in original graph
         # (after dead-end removed), then skip edges and return new graph as-is
         return H
 
     # STEP 6
     # create new edge from cluster to cluster for each edge in original graph
-    gdf_edges = utils_graph.graph_to_gdfs(G, nodes=False).set_index(["u", "v", "key"])
+    gdf_edges = utils_graph.graph_to_gdfs(G, nodes=False)
     for u, v, k, data in G.edges(keys=True, data=True):
         u2 = gdf.loc[u, "cluster"]
         v2 = gdf.loc[v, "cluster"]
@@ -552,7 +580,6 @@ def _consolidate_intersections_rebuild_graph(G, tolerance=10, reconnect_edges=Tr
     # STEP 7
     # for every group of merged nodes with more than 1 node in it, extend the
     # edge geometries to reach the new node point
-    new_edges = utils_graph.graph_to_gdfs(H, nodes=False)
     for cluster_label, nodes_subset in groups:
 
         # but only if there were multiple nodes merged together,
@@ -565,10 +592,11 @@ def _consolidate_intersections_rebuild_graph(G, tolerance=10, reconnect_edges=Tr
             y = H.nodes[cluster_label]["y"]
             xy = [(x, y)]
 
-            # for each edge incident to this new merged node, update
-            # its geometry to extend to/from the new node's point coords
-            mask = (new_edges["u"] == cluster_label) | (new_edges["v"] == cluster_label)
-            for _, (u, v, k) in new_edges.loc[mask, ["u", "v", "key"]].iterrows():
+            # for each edge incident to this new merged node, update its
+            # geometry to extend to/from the new node's point coords
+            in_edges = set(H.in_edges(cluster_label, keys=True))
+            out_edges = set(H.out_edges(cluster_label, keys=True))
+            for u, v, k in in_edges | out_edges:
                 old_coords = list(H.edges[u, v, k]["geometry"].coords)
                 new_coords = xy + old_coords if cluster_label == u else old_coords + xy
                 new_geom = LineString(new_coords)
