@@ -1,4 +1,6 @@
-"""Calculate graph-theoretic network measures."""
+"""Calculate geometric and topological network measures."""
+
+import logging as lg
 
 import networkx as nx
 import numpy as np
@@ -12,7 +14,7 @@ from . import utils_graph
 
 def basic_stats(G, area=None, clean_intersects=False, tolerance=15, circuity_dist="gc"):
     """
-    Calculate basic descriptive metric and topological stats for a graph.
+    Calculate basic descriptive geometric and topological stats for a graph.
 
     For an unprojected lat-lng graph, tolerance and graph units should be in
     degrees, and circuity_dist should be 'gc'. For a projected graph,
@@ -40,23 +42,23 @@ def basic_stats(G, area=None, clean_intersects=False, tolerance=15, circuity_dis
     Returns
     -------
     stats : dict
-        dictionary of network measures containing the following elements (some
-        keys may not be present, based on the arguments passed into the function):
+        network measures containing the following elements (some keys may not
+        be present, based on the arguments passed into the function):
 
           - n = number of nodes in the graph
           - m = number of edges in the graph
           - k_avg = average node degree of the graph
           - intersection_count = number of intersections in graph, that is,
-                nodes with >1 street emanating from them
-          - streets_per_node_avg = how many streets (edges in the undirected
-                representation of the graph) emanate from each node (ie,
-                intersection or dead-end) on average (mean)
-          - streets_per_node_counts = dict, with keys of number of streets
-                emanating from the node, and values of number of nodes with this
-                count
+                nodes with >1 physical street connected to them
+          - streets_per_node_avg = how many physical streets (edges in the
+                undirected representation of the graph) connect to each node
+                (ie, intersection or dead-end) on average (mean)
+          - streets_per_node_counts = dict with keys of number of physical
+                streets connecting to a node, and values of number of nodes
+                with this count
           - streets_per_node_proportion = dict, same as previous, but as a
                 proportion of the total, rather than counts
-          - edge_length_total = sum of all edge lengths in the graph, in meters
+          - edge_length_total = sum of all edge lengths in graph, in meters
           - edge_length_avg = mean edge length in the graph, in meters
           - street_length_total = sum of all edges in the undirected
                 representation of the graph
@@ -74,39 +76,30 @@ def basic_stats(G, area=None, clean_intersects=False, tolerance=15, circuity_dis
           - circuity_avg = edge_length_total divided by the sum of the great
                 circle distances between the nodes of each edge
           - self_loop_proportion = proportion of edges that have a single node
-                as its two endpoints (ie, the edge links nodes u and v, and u==v)
+                as its endpoints (ie, the edge links nodes u and v, and u==v)
           - clean_intersection_count = number of intersections in street
                 network, merging complex ones into single points
-          - clean_intersection_density_km = clean_intersection_count divided by
-                area in square kilometers
+          - clean_intersection_density_km = clean_intersection_count divided
+                by area in square kilometers
     """
     sq_m_in_sq_km = 1e6  # there are 1 million sq meters in 1 sq km
     Gu = None
+    node_ids = set(G.nodes)
 
-    # calculate the number of nodes, n, and the number of edges, m, in the graph
+    # calculate number of nodes, n, and number of edges, m, in the graph
     n = len(G)
     m = len(G.edges)
 
     # calculate the average degree of the graph
     k_avg = 2 * m / n
 
-    if "streets_per_node" in G.graph:
-        # get the degrees saved as a graph attribute (from an undirected
-        # representation of the graph). this is not the degree of the nodes in
-        # the directed graph, but rather represents the number of streets
-        # (unidirected edges) emanating from each node. see
-        # count_streets_per_node function.
-        spn = G.graph["streets_per_node"]
-    else:
-        # count how many street segments emanate from each node in this graph
-        spn = utils_graph.count_streets_per_node(G)
+    # get number of physical streets (undirected edges) connected to each node
+    spn = nx.get_node_attributes(G, "street_count")
+    if set(spn) != node_ids:
+        utils.log("Graph nodes changed since `street_count`s were calculated", level=lg.WARN)
 
-    # count number of intersections in graph, as nodes with >1 street emanating
-    # from them
-    node_ids = set(G.nodes())
-    intersection_count = len(
-        [True for node, count in spn.items() if (count > 1) and (node in node_ids)]
-    )
+    # count number of intersections in graph, as nodes with >1 street
+    intersect_count = sum(count > 1 and node in node_ids for node, count in spn.items())
 
     # calculate streets-per-node average: the average number of streets
     # (unidirected edges) incident to each node
@@ -123,23 +116,23 @@ def basic_stats(G, area=None, clean_intersects=False, tolerance=15, circuity_dis
     spnp = {num: count / n for num, count in spnc.items()}
 
     # calculate the total and average edge lengths
-    edge_length_total = sum([d["length"] for u, v, d in G.edges(data=True)])
+    edge_length_total = sum(d["length"] for u, v, d in G.edges(data=True))
     edge_length_avg = edge_length_total / m
 
-    # calculate the total and average street segment lengths (so, edges without
+    # calculate total and average street segment lengths (so, edges without
     # double-counting two-way streets)
     if Gu is None:
         Gu = utils_graph.get_undirected(G)
-    street_length_total = sum([d["length"] for u, v, d in Gu.edges(data=True)])
+    street_length_total = sum(d["length"] for u, v, d in Gu.edges(data=True))
     street_segments_count = len(Gu.edges(keys=True))
     street_length_avg = street_length_total / street_segments_count
 
     # calculate clean intersection counts
     if clean_intersects:
         points = simplification.consolidate_intersections(G, tolerance, False, False)
-        clean_intersection_count = len(points)
+        clean_intersect_count = len(points)
     else:
-        clean_intersection_count = None
+        clean_intersect_count = None
 
     # we can calculate density metrics only if area is not null
     if area is not None:
@@ -148,9 +141,9 @@ def basic_stats(G, area=None, clean_intersects=False, tolerance=15, circuity_dis
         # calculate node density as nodes per sq km
         node_density_km = n / area_km
 
-        # calculate intersection density as nodes with >1 street emanating from
-        # them, per sq km
-        intersection_density_km = intersection_count / area_km
+        # calculate intersection density as nodes with >1 physical street
+        # connected to them, per sq km
+        intersect_density_km = intersect_count / area_km
 
         # calculate edge density as linear meters per sq km
         edge_density_km = edge_length_total / area_km
@@ -159,26 +152,23 @@ def basic_stats(G, area=None, clean_intersects=False, tolerance=15, circuity_dis
         street_density_km = street_length_total / area_km
 
         if clean_intersects:
-            clean_intersection_density_km = clean_intersection_count / area_km
+            clean_intersect_density_km = clean_intersect_count / area_km
         else:
-            clean_intersection_density_km = None
+            clean_intersect_density_km = None
     else:
         # if area is None, then we cannot calculate density
         node_density_km = None
-        intersection_density_km = None
+        intersect_density_km = None
         edge_density_km = None
         street_density_km = None
-        clean_intersection_density_km = None
+        clean_intersect_density_km = None
 
     # average circuity: sum of edge lengths divided by sum of straight-line
     # distance between edge endpoints. first load all the edges origin and
     # destination coordinates as a dataframe, then calculate the straight-line
     # distance
-    coords = np.array(
-        [
-            [G.nodes[u]["y"], G.nodes[u]["x"], G.nodes[v]["y"], G.nodes[v]["x"]]
-            for u, v, k in G.edges(keys=True)
-        ]
+    coords = (
+        (G.nodes[u]["y"], G.nodes[u]["x"], G.nodes[v]["y"], G.nodes[v]["x"]) for u, v, k in G.edges
     )
     df_coords = pd.DataFrame(coords, columns=["u_y", "u_x", "v_y", "v_x"])
     if circuity_dist == "gc":
@@ -193,7 +183,7 @@ def basic_stats(G, area=None, clean_intersects=False, tolerance=15, circuity_dis
             y1=df_coords["u_y"], x1=df_coords["u_x"], y2=df_coords["v_y"], x2=df_coords["v_x"]
         )
     else:
-        raise ValueError('circuity_dist must be "gc" or "euclidean"')
+        raise ValueError('circuity_dist argument must be "gc" or "euclidean"')
 
     gc_distances = gc_distances.fillna(value=0)
     try:
@@ -201,17 +191,15 @@ def basic_stats(G, area=None, clean_intersects=False, tolerance=15, circuity_dis
     except ZeroDivisionError:
         circuity_avg = np.nan
 
-    # percent of edges that are self-loops, ie both endpoints are the same node
-    self_loops = [True for u, v, k in G.edges(keys=True) if u == v]
-    self_loops_count = len(self_loops)
-    self_loop_proportion = self_loops_count / m
+    # percent of edges that are self-loops, ie both endpoints are same node
+    self_loop_proportion = sum(u == v for u, v, k in G.edges) / m
 
     # assemble the results
     stats = {
         "n": n,
         "m": m,
         "k_avg": k_avg,
-        "intersection_count": intersection_count,
+        "intersection_count": intersect_count,
         "streets_per_node_avg": spna,
         "streets_per_node_counts": spnc,
         "streets_per_node_proportion": spnp,
@@ -221,13 +209,13 @@ def basic_stats(G, area=None, clean_intersects=False, tolerance=15, circuity_dis
         "street_length_avg": street_length_avg,
         "street_segments_count": street_segments_count,
         "node_density_km": node_density_km,
-        "intersection_density_km": intersection_density_km,
+        "intersection_density_km": intersect_density_km,
         "edge_density_km": edge_density_km,
         "street_density_km": street_density_km,
         "circuity_avg": circuity_avg,
         "self_loop_proportion": self_loop_proportion,
-        "clean_intersection_count": clean_intersection_count,
-        "clean_intersection_density_km": clean_intersection_density_km,
+        "clean_intersection_count": clean_intersect_count,
+        "clean_intersection_density_km": clean_intersect_density_km,
     }
 
     # return the results
@@ -236,7 +224,7 @@ def basic_stats(G, area=None, clean_intersects=False, tolerance=15, circuity_dis
 
 def extended_stats(G, connectivity=False, anc=False, ecc=False, bc=False, cc=False):
     """
-    Calculate extended topological stats and metrics for a graph.
+    Calculate extended topological measures for a graph.
 
     Many of these algorithms have an inherently high time complexity. Global
     topological analysis of large complex networks is extremely time consuming
@@ -307,7 +295,7 @@ def extended_stats(G, connectivity=False, anc=False, ecc=False, bc=False, cc=Fal
     # strongly connected graphs
     Gs = utils_graph.get_largest_component(G, strongly=True)
 
-    # average degree of the neighborhood of each node, and average for the graph
+    # average degree of the neighborhood of each node, and average for graph
     avg_neighbor_degree = nx.average_neighbor_degree(G)
     stats["avg_neighbor_degree"] = avg_neighbor_degree
     stats["avg_neighbor_degree_avg"] = sum(avg_neighbor_degree.values()) / len(avg_neighbor_degree)
@@ -352,11 +340,11 @@ def extended_stats(G, connectivity=False, anc=False, ecc=False, bc=False, cc=Fal
     # if True, calculate node and edge connectivity
     if connectivity:
 
-        # node connectivity is the minimum number of nodes that must be removed
+        # node connectivity is minimum number of nodes that must be removed
         # to disconnect G or render it trivial
         stats["node_connectivity"] = nx.node_connectivity(Gs)
 
-        # edge connectivity is equal to the minimum number of edges that must be
+        # edge connectivity is equal to minimum number of edges that must be
         # removed to disconnect G or render it trivial
         stats["edge_connectivity"] = nx.edge_connectivity(Gs)
         utils.log("Calculated node and edge connectivity")
@@ -364,7 +352,7 @@ def extended_stats(G, connectivity=False, anc=False, ecc=False, bc=False, cc=Fal
     # if True, calculate average node connectivity
     if anc:
         # mean number of internally node-disjoint paths between each pair of
-        # nodes in G, i.e., the expected number of nodes that must be removed to
+        # nodes in G, i.e., expected number of nodes that must be removed to
         # disconnect a randomly selected pair of non-adjacent nodes
         stats["node_connectivity_avg"] = nx.average_node_connectivity(G)
         utils.log("Calculated average node connectivity")
@@ -374,10 +362,8 @@ def extended_stats(G, connectivity=False, anc=False, ecc=False, bc=False, cc=Fal
     if ecc:
         # precompute shortest paths between all nodes for eccentricity-based
         # stats
-        sp = {
-            source: dict(nx.single_source_dijkstra_path_length(Gs, source, weight="length"))
-            for source in Gs.nodes()
-        }
+        length_func = nx.single_source_dijkstra_path_length
+        sp = {source: dict(length_func(Gs, source, weight="length")) for source in Gs.nodes}
 
         utils.log("Calculated shortest path lengths")
 
@@ -398,7 +384,7 @@ def extended_stats(G, connectivity=False, anc=False, ecc=False, bc=False, cc=Fal
         center = nx.center(Gs, e=eccentricity)
         stats["center"] = center
 
-        # periphery is the set of nodes with eccentricity equal to the diameter
+        # periphery is the set of nodes with eccentricity equal to diameter
         periphery = nx.periphery(Gs, e=eccentricity)
         stats["periphery"] = periphery
 
@@ -406,23 +392,19 @@ def extended_stats(G, connectivity=False, anc=False, ecc=False, bc=False, cc=Fal
     if cc:
         # closeness centrality of a node is the reciprocal of the sum of the
         # shortest path distances from u to all other nodes
-        closeness_centrality = nx.closeness_centrality(G, distance="length")
-        stats["closeness_centrality"] = closeness_centrality
-        stats["closeness_centrality_avg"] = sum(closeness_centrality.values()) / len(
-            closeness_centrality
-        )
+        close_cent = nx.closeness_centrality(G, distance="length")
+        stats["closeness_centrality"] = close_cent
+        stats["closeness_centrality_avg"] = sum(close_cent.values()) / len(close_cent)
         utils.log("Calculated closeness centrality")
 
     # if True, calculate node betweenness centrality
     if bc:
         # betweenness centrality of a node is the sum of the fraction of
-        # all-pairs shortest paths that pass through node
-        # networkx 2.4+ implementation cannot run on Multi(Di)Graphs, so use DiGraph
-        betweenness_centrality = nx.betweenness_centrality(D, weight="length")
-        stats["betweenness_centrality"] = betweenness_centrality
-        stats["betweenness_centrality_avg"] = sum(betweenness_centrality.values()) / len(
-            betweenness_centrality
-        )
+        # all-pairs shortest paths that pass through node. nx2.4+
+        # implementation cannot run on Multi(Di)Graphs, so use DiGraph
+        btwn_cent = nx.betweenness_centrality(D, weight="length")
+        stats["betweenness_centrality"] = btwn_cent
+        stats["betweenness_centrality_avg"] = sum(btwn_cent.values()) / len(btwn_cent)
         utils.log("Calculated betweenness centrality")
 
     utils.log("Calculated extended stats")
